/*
 * kgraft_patch_bsc1173666
 *
 * Fix for CVE-2019-14816, bsc#1173666
 *
 *  Upstream commit:
 *  7caac62ed598 ("mwifiex: Fix three heap overflow at parsing element in
 *                 cfg80211_ap_settings")
 *
 *  SLE12-SP2 and -SP3 commit:
 *  b6220dc818be0869388f09d59affbff30f192872
 *
 *  SLE12-SP4, SLE12-SP5, SLE15 and SLE15-SP1 commit:
 *  3d20ec501d456b6f9ae8d84d16b107216ac47664
 *
 *
 *  Copyright (c) 2020 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_MWIFIEX)

#if !IS_MODULE(CONFIG_MWIFIEX)
#error "Live patch supports only CONFIG_MWIFIEX=m"
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include "kgr_patch_bsc1173666.h"
#include "../kallsyms_relocs.h"

#define KGR_PATCHED_MODULE "mwifiex"


/* klp-ccp: from drivers/net/wireless/mwifiex/main.h */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/ip.h>
#include <linux/skbuff.h>

/* klp-ccp: from drivers/net/wireless/mwifiex/main.h */
#include <linux/of.h>
#include <linux/idr.h>
/* klp-ccp: from drivers/net/wireless/mwifiex/decl.h */
#include <linux/wait.h>
#include <linux/timer.h>
#include <linux/ieee80211.h>
#include <uapi/linux/if_arp.h>

/* klp-ccp: from include/net/cfg80211.h */
static const u8 *(*kgre_cfg80211_find_vendor_ie)(unsigned int oui, u8 oui_type,
				  const u8 *ies, int len);

/* klp-ccp: from drivers/net/wireless/mwifiex/ioctl.h */
#include <net/lib80211.h>

#define IEEE_MAX_IE_SIZE		256

/* klp-ccp: from drivers/net/wireless/mwifiex/fw.h */
#include <linux/if_ether.h>

#define MWIFIEX_AUTO_IDX_MASK			0xffff

struct mwifiex_ie {
	__le16 ie_index;
	__le16 mgmt_subtype_mask;
	__le16 ie_length;
	u8 ie_buffer[IEEE_MAX_IE_SIZE];
} __packed;

/* klp-ccp: from drivers/net/wireless/mwifiex/pcie.h */
#include    <linux/interrupt.h>
/* klp-ccp: from drivers/net/wireless/mwifiex/sdio.h */
#include <linux/mmc/sdio.h>
#include <linux/mmc/sdio_ids.h>

/* klp-ccp: from drivers/net/wireless/mwifiex/main.h */
struct ieee_types_header {
	u8 element_id;
	u8 len;
} __packed;

/* klp-ccp: from drivers/net/wireless/mwifiex/ie.c */
int kgrp_mwifiex_update_vs_ie(const u8 *ies, int ies_len,
				struct mwifiex_ie **ie_ptr, u16 mask,
				unsigned int oui, u8 oui_type)
{
	struct ieee_types_header *vs_ie;
	struct mwifiex_ie *ie = *ie_ptr;
	const u8 *vendor_ie;

	vendor_ie = (*kgre_cfg80211_find_vendor_ie)(oui, oui_type, ies, ies_len);
	if (vendor_ie) {
		if (!*ie_ptr) {
			*ie_ptr = kzalloc(sizeof(struct mwifiex_ie),
					  GFP_KERNEL);
			if (!*ie_ptr)
				return -ENOMEM;
			ie = *ie_ptr;
		}

		vs_ie = (struct ieee_types_header *)vendor_ie;
		/*
		 * Fix CVE-2019-14816
		 *  +3 lines
		 */
		if (le16_to_cpu(ie->ie_length) + vs_ie->len + 2 >
			IEEE_MAX_IE_SIZE)
			return -EINVAL;
		memcpy(ie->ie_buffer + le16_to_cpu(ie->ie_length),
		       vs_ie, vs_ie->len + 2);
		le16_add_cpu(&ie->ie_length, vs_ie->len + 2);
		ie->mgmt_subtype_mask = cpu_to_le16(mask);
		ie->ie_index = cpu_to_le16(MWIFIEX_AUTO_IDX_MASK);
	}

	*ie_ptr = ie;
	return 0;
}



static struct kgr_kallsyms_reloc kgr_funcs[] = {
	{ "cfg80211_find_vendor_ie", (void *)&kgre_cfg80211_find_vendor_ie,
	  "cfg80211" },
};

static int kgr_patch_bsc1173666_module_notify(struct notifier_block *nb,
					      unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = __kgr_resolve_kallsyms_relocs(kgr_funcs, ARRAY_SIZE(kgr_funcs));
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1173666_module_nb = {
	.notifier_call = kgr_patch_bsc1173666_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1173666_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = __kgr_resolve_kallsyms_relocs(kgr_funcs,
						    ARRAY_SIZE(kgr_funcs));
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1173666_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1173666_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1173666_module_nb);
}

#endif /* IS_ENABLED(CONFIG_MWIFIEX) */
